/*
  Copyright (C) 2003 Fons Adriaensen
    
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <math.h>
#include "vco_sawpulse.h"


extern float exp2ap (float x);


static float _pulse [] = 
{
    0.000000,  0.000366,  0.001458,  0.003117,  0.005028,  0.006756,  0.007806,  0.007704,
    0.006086,  0.002778, -0.002139, -0.008292, -0.015022, -0.021434, -0.026492, -0.029153,
   -0.028511, -0.023957, -0.015297, -0.002869,  0.012417,  0.029105,  0.045302,  0.058853,
    0.067578,  0.069533,  0.063273,  0.048113,  0.024337, -0.006682, -0.042487, -0.079654,
   -0.114036, -0.141088, -0.156243, -0.155364, -0.135210, -0.093843, -0.030932,  0.052092,
    0.152049,  0.264172,  0.382359,  0.499585,  0.608450,  0.701808,  0.773391,  0.818353,
    0.833670,  0.818353,  0.773391,  0.701808,  0.608450,  0.499585,  0.382359,  0.264172,
    0.152049,  0.052092, -0.030932, -0.093843, -0.135210, -0.155364, -0.156243, -0.141088,
   -0.114036, -0.079654, -0.042487, -0.006682,  0.024337,  0.048113,  0.063273,  0.069533,
    0.067578,  0.058853,  0.045302,  0.029105,  0.012417, -0.002869, -0.015297, -0.023957,
   -0.028511, -0.029153, -0.026492, -0.021434, -0.015022, -0.008292, -0.002139,  0.002778,
    0.006086,  0.007704,  0.007806,  0.006756,  0.005028,  0.003117,  0.001458,  0.000366,
    0.000000
};



void Ladspa_VCO_pulse1::setport (unsigned long port, LADSPA_Data *data)
{
    _port [port] = data;
}


void Ladspa_VCO_pulse1::active (bool act)
{
    _p = 0.5;
    _w = _y = _z = _r0 = _r1  = 0;
    _i0 = 0;
    _i1 = 0;
}


void Ladspa_VCO_pulse1::runproc (unsigned long len, bool add)
{
    int    i0, i1, k;
    float  *outp, *freq, *expm, *linm;   // pointers to audio rate port data
    float  p, w, y, z, r0, r1;
    float  a, t, dw;

    outp = _port[OUTP];
    freq = _port[FREQ] - 1;
    expm = _port[EXPM] - 1;
    linm = _port[LINM] - 1;
    
    p = _p;
    w = _w;
    y = _y;
    z = _z;
    r0 = _r0;
    r1 = _r1;
    i0 = _i0;
    i1 = _i1;

    do
    {
        k = (len > 24) ? 16 : len;
        freq += k;
        expm += k;
        linm += k;
        len -= k;
 
        t = (exp2ap (*freq + _port[OCTN][0] + _port[TUNE][0] + *expm * _port[EXPG][0] + 8.03136)
	     + 1e3 * *linm * _port[LING][0]) / _fsam; 
        if (t < 1e-5) t = 1e-5;
        if (t > 0.165) t = 0.165;
        dw = (t - w) / k;
        a = 0.2 + 0.8 * _port [FILT][0];

        while (k--)
	{
	    w += dw; 
            p += w;
            if (p >= 1.0)
	    {   
		p -= 1.0;
                if (i0 > 0)
		{
		    r1 = r0;
                    i1 = i0;
		}
                r0 = NPHASE * p / w;
                i0 = (int) r0;
                r0 -= i0;
                i0 = NPHASE * NCOEFF - i0;
	    }
            y = 0;
            if (i0 > 0)
	    {
		y += r0 * _pulse [i0 - 1 ] + (1 - r0) * _pulse [i0];
                i0 -= NPHASE;
	    } 
            if (i1 > 0)
	    {
		y += r1 * _pulse [i1 - 1 ] + (1 - r1) * _pulse [i1];
                i1 -= NPHASE;
	    } 
            z += a * (y - z);
            *outp++ = z;
	}
    }
    while (len);

    _p = p;
    _w = w;
    _y = y;
    _z = z;
    _r0 = r0;
    _r1 = r1;
    _i0 = i0;
    _i1 = i1;
}



void Ladspa_VCO_saw1::setport (unsigned long port, LADSPA_Data *data)
{
    _port [port] = data;
}


void Ladspa_VCO_saw1::active (bool act)
{
    _p = 0.5;
    _w = _x = _y = _z = _d = _r0 = _r1  = 0;
    _i0 = 0;
    _i1 = 0;
}


void Ladspa_VCO_saw1::runproc (unsigned long len, bool add)
{
    int    i0, i1, k;
    float  *outp, *freq, *expm, *linm, *sync;   // pointers to audio rate port data
    float  p, w, x, y, z, d, r0, r1;
    float  a, t, dw;

    outp = _port[OUTP];
    freq = _port[FREQ] - 1;
    expm = _port[EXPM] - 1;
    linm = _port[LINM] - 1;
    sync = _port[SYNC];

    p = _p;
    w = _w;
    x = _x;
    y = _y;
    z = _z;
    d = _d;
    r0 = _r0;
    r1 = _r1;
    i0 = _i0;
    i1 = _i1;

    do
    {
        k = (len > 24) ? 16 : len;
        freq += k;
        expm += k;
        linm += k;
        len -= k;
 
        t = (exp2ap (*freq + _port[OCTN][0] + _port[TUNE][0] + *expm * _port[EXPG][0] + 8.03136 + d)
	     + 1e3 * *linm * _port[LING][0]) / _fsam; 
        if (t < 1e-5) t = 1e-5;
        if (t > 0.165) t = 0.165;
        dw = (t - w) / k;
        a = 0.2 + 0.8 * _port [FILT][0];

        while (k--)
	{
	    w += dw; 
            x -= w * (0.01 * y + 0.1 * x + 1);
            p += w;
            if (p >= 1.0)
	    {   
		p -= 1.0;
                if (i0 > 0)
		{
		    r1 = r0;
                    i1 = i0;
		}
                r0 = NPHASE * p / w;
                i0 = (int) r0;
                r0 -= i0;
                i0 = NPHASE * NCOEFF - i0;
	    }
            if (i0 > 0)
	    {
		x += r0 * _pulse [i0 - 1 ] + (1 - r0) * _pulse [i0];
                i0 -= NPHASE;
	    } 
            if (i1 > 0)
	    {
		x += r1 * _pulse [i1 - 1 ] + (1 - r1) * _pulse [i1];
                i1 -= NPHASE;
	    } 
            y += 6.3 * w * x;
            z += a * (x - z);
            *outp++ = z;
            d += 0.01 * (y * *sync++ - d);
	}
    }
    while (len);

    _p = p;
    _w = w;
    _x = x;
    _y = y;
    _z = z;
    _d = d;
    _r0 = r0;
    _r1 = r1;
    _i0 = i0;
    _i1 = i1;
}
